import argparse
import configparser
import pathlib
import shlex
import subprocess
import webbrowser

from . import kattis
from . import util


def cli():
    """Enter CLI here."""
    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument(
        "--config",
        default=util.nearest(".kattisrc"),
        help="config file (defaults to nearest '.kattisrc')",
    )

    parser = argparse.ArgumentParser(description="")
    parser.set_defaults(callback=nothing)

    subparsers = parser.add_subparsers(metavar="{init | submit | test}")

    init_parser = subparsers.add_parser("init", parents=[parent])
    init_parser.add_argument("problem", help="Kattis problem ID")
    init_parser.set_defaults(callback=init)

    submit_parser = subparsers.add_parser("submit", parents=[parent])
    submit_parser.add_argument("--force", "-f", help="force submission")
    submit_parser.set_defaults(callback=submit)

    test_parser = subparsers.add_parser("test", parents=[parent])
    test_parser.set_defaults(callback=test)

    args = parser.parse_args()
    args.callback(**vars(args))


def _kati_config():
    config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation()
    )

    if not pathlib.Path("kati").is_file():
        raise FileNotFoundError("kati")
    config.read("kati")

    return config


def init(config, problem, **kwargs):
    problem = kattis.Problem(config, problem)
    kati_config = _kati_config()
    kati_config["kattis"]["problem"] = problem.id

    with problem.samples() as samples:
        dest = pathlib.Path(kati_config["kattis"]["samples"])
        dest.mkdir(exist_ok=True)
        samples.extractall(dest)

    with open("kati", mode="w") as file_:
        kati_config.write(file_)


def nothing(**kwargs):
    pass


def submit(config, force, **kwargs):
    if force or input("submit (y/n) ") == "y":
        kati_config = _kati_config()
        kati_config[kati_config.default_section]["main"] = ""

        problem = kattis.Problem(config, kati_config["kattis"]["problem"])
        verdict = problem.submit(
            [pathlib.Path(file_) for file_ in kati_config["kattis"]["files"].split(" ")],
            kati_config["kattis"]["language"],
            kati_config["kattis"]["main"],
        )
        print(verdict)
        webbrowser.open(verdict)


def test(config, **kwargs):
    kati_config = _kati_config()
    for file_ in pathlib.Path(kati_config["kattis"]["samples"]).iterdir():
        if file_.suffix == ".in":
            print(f"... test '{file_}'")
            _test(
                kati_config, file_.read_bytes(), file_.with_suffix(".ans").read_bytes()
            )


def _test(kati_config, input_, output):
    try:
        if "build" in kati_config["kattis"]:
            print("\tbuilding...")
            proc = subprocess.run(
                shlex.split(kati_config["kattis"]["build"]),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            proc.check_returncode()

        proc = subprocess.run(
            shlex.split(kati_config["kattis"]["run"]),
            input=input_,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        proc.check_returncode()

        if proc.stdout == output:
            print("\tpassed")
        else:
            print("\twrong answer")
            print("\tinput:\n{}\n".format(input_.decode("UTF-8")))
            print("\texpected:\n{}\n".format(output.decode("UTF-8")))
            print("\tfound:\n{}\n".format(proc.stdout.decode("UTF-8")))
    except subprocess.CalledProcessError as err:
        print("\truntime error")
        print(proc.stdout)
        raise
