import configparser
import collections
import io
import pathlib
import re
import requests
import requests.exceptions
import urllib.parse
import zipfile


class Error(Exception):
    pass


class ConfigError(Error):
    """Raise when config file is bad."""


class LoginError(Error):
    """Raise when a login request to Kattis fails."""


class SubmissionError(Error):
    """Raise when a submission to Kattis fails."""


class SamplesError(Error):
    """Raise when an error occurs while downloading or extracting a test case."""


class UnsupportedLanguageError(ValueError):
    """Raise when Kattis does not support a submission's language."""


class Problem:
    """Store, retrieve, and interact with a Kattis problem.

    All interaction with Kattis should be done through this class.
    """

    _headers = {"User-Agent": "kattis-cli-submit"}
    _languages = [
        "C",
        "C++",
        "C#",
        "COBOL",
        "F#",
        "Go",
        "Haskell",
        "Java",
        "JavaScript",
        "Kotlin",
        "Common Lisp",
        "Objective-C",
        "OCaml",
        "Pascal",
        "PHP",
        "Prolog",
        "Python 2",
        "Python 3",
        "Ruby",
        "Rust",
        "Scala",
    ]

    def __init__(self, config, id_):
        """Download a Kattis problem.

        :param config: a path-like object pointing to config file
        :param id_: Kattis problem ID
        """
        self.config = config
        self.id = id_

        if not self.config.is_file():
            raise FileNotFoundError(self.config)
        self.parser = configparser.ConfigParser()
        self.parser.read(self.config)

    def samples(self):
        """Download sample test cases.

        :return: a `zipfile.ZipFile` object containing the suite of test cases
        """
        try:
            response = requests.get(
                "https://{}/problems/{}/file/statement/samples.zip".format(
                    self.parser["kattis"]["hostname"], self.id
                ),
                headers=self._headers,
            )
            response.raise_for_status()

            return zipfile.ZipFile(io.BytesIO(response.content))
        except requests.exceptions.RequestException as err:
            raise SamplesError() from err

    def _login(self):
        """Login to Kattis.

        :return: freshly baked cookies
        """
        data = {"user": self.parser["user"]["username"], "script": "true"}

        # Config must either provide...
        if "password" in self.parser["user"]:
            # a password or...
            data["password"] = self.parser["user"]["password"]
        elif "token" in self.parser["user"]:
            # a token.
            data["token"] = self.parser["user"]["token"]
        else:
            raise ConfigError(self.config)

        try:
            response = requests.post(
                self.parser["kattis"]["loginurl"], data=data, headers=self._headers,
            )
            response.raise_for_status()
            return response.cookies
        except requests.exceptions.RequestException as err:
            raise LoginError(response.status_code) from err

    def submit(self, files, language, main_class=""):
        """Submit a solution to this problem.

        :param files: a list of path-like objects
        :param language:
        :param main_class:
        :return: a URL to the submission page where the results of the solution's judgement may be viewed
        """

        if language not in self._languages:
            raise UnsupportedLanguageError(language)

        cookies = self._login()

        data = {
            "submit": "true",
            "submit_ctr": 2,
            "language": language,
            "mainclass": main_class,
            "problem": self.id,
            "tag": "",
            "script": "true",
        }
        files = [
            ("sub_file[]", (file.name, file.read_bytes(), "application/octet-stream"))
            for file in files
        ]

        try:
            response = requests.post(
                self.parser["kattis"]["submissionurl"],
                data=data,
                files=files,
                cookies=cookies,
                headers=self._headers,
            )
            response.raise_for_status()
            content = response.content.decode("UTF-8")
            match = re.search(r"Submission ID: (\d+)", content)
            if match:
                return "/".join(
                    [self.parser["kattis"]["submissionsurl"], match.group(1)]
                )
            else:
                # If the response does not contain a submission ID,
                # it is almost certainly not good news.
                raise SubmissionError(content)
        except requests.exceptions.RequestException as err:
            raise SubmissionError() from err

Test = collections.namedtuple("Test", ["input", "answer"])
Test.__doc__ = "Represent a test case as a named tuple (input, answer)."
